/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */

/*
 *    Misc Brutus specific definitions. 
 *
 *    Authors:
 *        Jules Colding <colding@42tools.com>
 *
 *    Copyright (C) 2005-2008 OMC Denmark ApS.
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, 
 *    MA 02111-1307 USA
 */

#ifndef BRUTUS_H
#define BRUTUS_H 1

#include <glib.h>
#include <glib/gmacros.h>
#include <execinfo.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>

#include <idl_output/IMAPISession.h>
#include <session/brutus-proxy.h>
#include <session/brutusd-util.h>
#include <server/brutus_string.h>

G_BEGIN_DECLS

/*************************
 *       Typedefs        *
 ************************/

typedef guchar digest_t[16];    // MD5 digest
typedef gchar digest_str_t[33]; // stringified version of digest
typedef	char uuid_str_t[37];    // stringified UUID

typedef struct _BrutusBaseClass {
	CORBA_ORB orb;                       // ORB
	PortableServer_POA root_poa;         // POA

	BRUTUS_IMAPISession mapi_session;    // store-global IMAPISession
	BRUTUS_BrutusProxy brutus_proxy;     // IPv4/6 capable Brutus proxy
	gboolean expect_other_clients;       // do we expect that other clients are accessing the same mailbox?
	gboolean likely_connected;           // best-effort estimate on connectivity status
	GStaticMutex likely_connected_mutex;

	struct BrutusCheckProxy lifeline;    // lifeline callback, local and proxy

	CORBA_unsigned_long instance_id;     // IMAPISession instance id
	long exchange_time_zone;             // The Exchange Server timezone offset from GMT in minutes 
	long client_time_zone;               // The evolution-brutus timezone offset from GMT in minutes 

	GMutex *connect_mutex;               // protecting (re-)connect attempts

	CORBA_char *mapi_profile;
	gchar *brutus_logon_str;             // corbaloc IOR of remote BrutusLogOn object
	gchar *cache_dir;                    // cache directory - camel storage_path
	gchar *brutus_server;                // Brutus Server box
	gchar *exchange_server;              // Exchange server box

	guint16 brutus_port;                 // ListenEndpoint port number of remote BrutusLogOn object, Lorica mapped or not
} BrutusBaseClass;

#define BRUTUS_BASE_CLASS(_bc) ((BrutusBaseClass*)_bc)

/*************************
 *      Capabilities     *
 ************************/

/* create the file $HOME/.brutus-debug if you want debugging output - delete it if you don't. It may or may not be a regular file */
#define BRUTUS_DO_DEBUG_FILE_NAME ".brutus-do-debug"

/* Prepend before all output */
#define DEBUG_PREFIX "\n** BRUTUS "PACKAGE_VERSION

/* Do we support SSLIOP? */
#define BRUTUS_HAVE_SSL (1)
#undef BRUTUS_HAVE_SSL // not tested yet

/* Do we support encrypted passwords? */
#define BRUTUS_HAVE_ENCRYPTED_AUTH (1)
#undef BRUTUS_HAVE_ENCRYPTED_AUTH  // not implemented yet

/*************************
 *       Debug macros    *
 ************************/

#ifdef BRUTUS_DEBUG

extern void brutus_DRAW(const char *dbg_msg);

extern void brutus_print_orbit2_ex(const int line,
				   const char *file,
				   const char *function,
				   const CORBA_Environment *ev);

extern void brutus_print_if_orbit2_ex(const int line,
				      const char *file,
				      const char *function,
				      const CORBA_Environment *ev);
extern void brutus_D(const int line,
		     const char *file,
		     const char *function,
		     const char *dbg_msg);

extern void brutus_DBT(const int line,
		       const char *file,
		       const char *function);

#endif

#undef ORBIT2_EX
#define ORBIT2_EX(_ev_) ((NULL != _ev_) && ((_ev_)->_major != CORBA_NO_EXCEPTION))

#undef PRINT_ORBIT2_EX
#ifdef BRUTUS_DEBUG
#define PRINT_ORBIT2_EX(_ev_)                                                     \
        G_STMT_START {                                                            \
		brutus_print_orbit2_ex(__LINE__, __FILE__, __FUNCTION__, (_ev_)); \
	} G_STMT_END
#else
#define PRINT_ORBIT2_EX(_ev_) G_STMT_START { } G_STMT_END
#endif

#undef PRINT_IF_ORBIT2_EX
#ifdef BRUTUS_DEBUG
#define PRINT_IF_ORBIT2_EX(_ev_)                                                     \
        G_STMT_START {                                                               \
		brutus_print_if_orbit2_ex(__LINE__, __FILE__, __FUNCTION__, (_ev_)); \
	} G_STMT_END
#else
#define PRINT_IF_ORBIT2_EX(_ev_) G_STMT_START { } G_STMT_END
#endif

#undef d
#ifdef BRUTUS_DEBUG
#define d(_str_)                                                   \
        G_STMT_START {                                             \
		brutus_D(__LINE__, __FILE__, __FUNCTION__, _str_); \
        } G_STMT_END
#else
#define d(_str_) G_STMT_START { } G_STMT_END
#endif

#undef dfmt
#ifdef BRUTUS_DEBUG
#define dfmt(_fmt_, _str_)                                             \
        G_STMT_START {                                                 \
		char *_dbg_str_ = brutus_to_string(_fmt_, _str_);      \
		brutus_D(__LINE__, __FILE__, __FUNCTION__, _dbg_str_); \
		free(_dbg_str_);                                       \
        } G_STMT_END
#else
#define dfmt(_fmt_, _str_) G_STMT_START { } G_STMT_END
#endif

#undef dfmt2
#ifdef BRUTUS_DEBUG
#define dfmt2(_fmt_, _str1_, _str2_)                                       \
        G_STMT_START {                                                     \
		char *_dbg_str_ = brutus_to_string(_fmt_, _str1_, _str2_); \
		brutus_D(__LINE__, __FILE__, __FUNCTION__, _dbg_str_);     \
		free(_dbg_str_);                                           \
        } G_STMT_END
#else
#define dfmt2(_fmt_, _str1_, _str2_) G_STMT_START { } G_STMT_END
#endif

#undef dbt
#ifdef BRUTUS_DEBUG
#define dbt()                                          \
        G_STMT_START {                                 \
		DBT(__LINE__, __FILE__, __FUNCTION__); \
        } G_STMT_END
#else
#define dbt() G_STMT_START { } G_STMT_END
#endif


/*************************
 *      Utility macros   *
 ************************/

#undef BRUTUS_NULL_IF_EMPTY_STRING
#define BRUTUS_NULL_IF_EMPTY_STRING(_str_) G_STMT_START { if (_str_ && !strlen(_str_)) _str_ = NULL; } G_STMT_END


#undef BRUTUS_CHECK_ALLOC
#define BRUTUS_CHECK_ALLOC(_mem_, _ex_)                                                              \
        G_STMT_START {                                                                               \
	        if (!_mem_) {                                                                        \
		        d("No memory");                                                              \
		        if (_ex_) camel_exception_set(_ex_, CAMEL_EXCEPTION_SYSTEM, _("No memory")); \
		        goto out;                                                                    \
	        }                                                                                    \
	} G_STMT_END


#undef BRUTUS_CLOSE_FD
#define BRUTUS_CLOSE_FD(__fd__)                                   \
        G_STMT_START {                                            \
		while ((-1 == close(__fd__)) && (errno == EINTR)) \
			;                                         \
	} G_STMT_END

/*************************
 *      GTimer macros    *
 ************************/

#undef BRUTUS_TIMER_NEW
#ifdef BRUTUS_DEBUG
#define BRUTUS_TIMER_NEW() GTimer *_brutus_timer_ = NULL; gdouble _elapsed_time_ = 0.0; char *_do_debug_file = g_strconcat(getenv("HOME"), "/", BRUTUS_DO_DEBUG_FILE_NAME, NULL);
#else
#define BRUTUS_TIMER_NEW() G_STMT_START { } G_STMT_END
#endif

#undef BRUTUS_TIMER_INIT
#ifdef BRUTUS_DEBUG
#define BRUTUS_TIMER_INIT()                                                \
        G_STMT_START {				                           \
		if (g_file_test(_do_debug_file, G_FILE_TEST_IS_REGULAR)) { \
			_brutus_timer_ = g_timer_new();                    \
			g_timer_stop(_brutus_timer_);	                   \
		}                                                          \
	} G_STMT_END
#else
#define BRUTUS_TIMER_INIT() G_STMT_START { } G_STMT_END
#endif

#undef BRUTUS_TIMER_FREE
#ifdef BRUTUS_DEBUG
#define BRUTUS_TIMER_FREE()                                                \
        G_STMT_START {				                           \
		if (g_file_test(_do_debug_file, G_FILE_TEST_IS_REGULAR)) { \
			g_timer_destroy(_brutus_timer_);                   \
			_elapsed_time_ = 0.0;                              \
		}                                                          \
		g_free(_do_debug_file);                                    \
	} G_STMT_END
#else
#define BRUTUS_TIMER_FREE() G_STMT_START { } G_STMT_END
#endif

#undef BRUTUS_TIMER_START
#ifdef BRUTUS_DEBUG
#define BRUTUS_TIMER_START()                                               \
        G_STMT_START {                                                     \
                if (g_file_test(_do_debug_file, G_FILE_TEST_IS_REGULAR)) { \
                        g_timer_start(_brutus_timer_);                     \
                        fflush(NULL);                                      \
                }                                                          \
        } G_STMT_END
#else
#define BRUTUS_TIMER_START() G_STMT_START { } G_STMT_END
#endif

#undef BRUTUS_TIMER_PRINT_ELAPSED
#ifdef BRUTUS_DEBUG
#define BRUTUS_TIMER_PRINT_ELAPSED()                                                                                                          \
        G_STMT_START {                                                                                                                        \
                if (g_file_test(_do_debug_file, G_FILE_TEST_IS_REGULAR)) {                                                                    \
                        g_timer_stop(_brutus_timer_);                                                                                         \
                        _elapsed_time_ = g_timer_elapsed(_brutus_timer_, NULL);                                                               \
                        g_print(DEBUG_PREFIX" %s() in %s@%d: Elapsed time = %f seconds\n", __FUNCTION__, __FILE__, __LINE__, _elapsed_time_); \
                        fflush(NULL);                                                                                                         \
                }                                                                                                                             \
        } G_STMT_END
#else
#define BRUTUS_TIMER_PRINT_ELAPSED() G_STMT_START { } G_STMT_END
#endif


G_END_DECLS

#endif /* BRUTUS_H */
