/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */

/*
 *    Servant implementation file for a Brutus proxy daemon.
 *    Copyright (C) 2005-2007 OMC Denmark ApS.
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *    MA 02111-1307 USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <session/brutus-proxy.h>
#include <session/servants/BrutusCheck_impl.c>
#include <session/servants/IMAPIAdviseSink_impl.c>
#include <session/servants/types_impl.c>

#include <stdio.h>

/* Lorica */
#include "ReferenceMapper.h"

#include "brutus-proxy.h"

struct WrapperItem {
	CORBA_unsigned_long key;
	CORBA_Object local_object;
};

struct MAPIAccount {
	CORBA_char *mailbox;
	CORBA_char *server;
	CORBA_char *profile;
	BRUTUS_IMAPISession session;
	CORBA_unsigned_long instance_id;
};

#ifndef ORBIT2_EX
#define ORBIT2_EX(_ev_) ((NULL != _ev_) && ((_ev_)->_major != CORBA_NO_EXCEPTION))
#endif

static GStaticMutex st_mutex = G_STATIC_MUTEX_INIT;
static GStaticMutex map_mutex = G_STATIC_MUTEX_INIT;

/*
 * RFC 3696
 */
static gboolean
brutus_valid_dns_name(const char *name)
{
	const char *tmp = name;

	if (!name || !strlen(name))
		return FALSE;

	if ('-' == *name)
		return FALSE;

	if ('-' == name[strlen(name)-1])
		return FALSE;

	while ( '\0' != *name) {
		if (!g_ascii_isalnum(*name)
		    && ('-' != *name)
		    && ('.' != *name))
			return FALSE;

		name++;
	}

	return (g_strstr_len(tmp, strlen(tmp), "..") ? FALSE : TRUE);
}

static gboolean
brutus_server_ping(const char *brutus_server)
{
	int sys_retv = -1;
	char *cmd = NULL;

	if (!brutus_valid_dns_name(brutus_server))
		return FALSE;

	cmd = g_strdup_printf("/bin/ping -c 1 %s &>/dev/null", brutus_server);
	if (!cmd)
		return FALSE;
	sys_retv = system(cmd);
	g_free(cmd);
	if (sys_retv)
		return FALSE;

	return TRUE;
}

static gboolean
brutus_is_connected(const char *brutus_server,
		    BRUTUS_IMAPISession mapi_session)
{
	gboolean retv = FALSE;
	CORBA_boolean cb = CORBA_TRUE;
	CORBA_Environment ev[1];

	if (!brutus_server || !strlen(brutus_server))
		return FALSE;

#ifndef HAVE_ORBIT2_TIMEOUT
	if (!brutus_server_ping(brutus_server))
		return FALSE;
#endif
	CORBA_exception_init(ev);

	cb = CORBA_Object_is_nil((CORBA_Object)mapi_session, ev);
	if (ORBIT2_EX(ev))
		goto out;
	if (!cb) {
		BRUTUS_IMAPISession_ping(mapi_session, ev);
		if (!ORBIT2_EX(ev))
			retv = TRUE; // we are connected
	}

out:
	CORBA_exception_free(ev);
	return retv;
}

static gboolean
brutus_account_find_match(const GPtrArray *accounts,
			  const CORBA_char *mailbox,
			  const CORBA_char *server,
			  guint *idx)
{
	guint n = 0;
	struct MAPIAccount *account = NULL;

	if (!accounts)
		return FALSE;

	for (n = 0; n < accounts->len; n++) {
		account = (struct MAPIAccount*)g_ptr_array_index(accounts, n);
		if (!account)
			continue;
		if (!strcmp(mailbox, account->mailbox)
		    && !strcmp(server, account->server)) {
			*idx = n;
			return TRUE;
		}
	}

	return FALSE;
}

static gboolean
brutus_account_find_session(GPtrArray *accounts,
			    const CORBA_char *mailbox,
			    const CORBA_char *server,
			    guint *idx)
{
	guint n = 0;
	struct MAPIAccount *account = NULL;
	CORBA_boolean cb = CORBA_TRUE;
	CORBA_Environment ev[1];

	if (!accounts)
		return FALSE;

	CORBA_exception_init(ev);

	for (n = 0; n < accounts->len; n++) {
		account = (struct MAPIAccount*)g_ptr_array_index(accounts, n);
		if (!account)
			continue;

		if (!strcmp(mailbox, account->mailbox) && !strcmp(server, account->server)) {
			if (brutus_is_connected(account->server, account->session)) {
				CORBA_exception_free(ev);
				*idx = n;
				return TRUE;
			} else {
				cb = CORBA_Object_is_nil((CORBA_Object)account->session, ev);
				if (ORBIT2_EX(ev)) {
					CORBA_exception_free(ev);
					account->session = CORBA_OBJECT_NIL;
					return FALSE;
				}
				if (!cb) {
					CORBA_Object_release(account->session, ev);
					account->session = CORBA_OBJECT_NIL;
				}
			}
			break;
		}
	}
	CORBA_exception_free(ev);

	return FALSE;
}

static void
brutus_free_accounts(GPtrArray *accounts)
{
	guint n = 0;
	struct MAPIAccount *account = NULL;
	CORBA_boolean cb = CORBA_TRUE;
	CORBA_Environment ev[1];

	if (!accounts)
		return;

	CORBA_exception_init(ev);

	for (n = 0; n < accounts->len; n++) {
		account = (struct MAPIAccount*)g_ptr_array_index(accounts, n);
		if (!account)
			continue;

		/* no need the _Destroy() the session as the server will kill it anyway when the lifeline is gone */
		cb = CORBA_Object_is_nil(account->session, ev);
		if (!ORBIT2_EX(ev) && !cb)
			CORBA_Object_release(account->session, ev);

		if (account->mailbox)
			CORBA_free(account->mailbox);
		if (account->server)
			CORBA_free(account->server);
		if (account->profile)
			CORBA_free(account->profile);
		g_free(account);
	}
	g_ptr_array_free(accounts, TRUE);

	CORBA_exception_free(ev);
}

/* Will write all accounts to disk */
static void
brutus_store_accounts(GPtrArray *accounts)
{
	guint n = 0;
	struct MAPIAccount *account = NULL;
	FILE *accounts_file = NULL;
	int accounts_file_fd = -1;
	gchar *accounts_file_path = NULL;

	accounts_file_path = g_strconcat(getenv("HOME"), "/", BRUTUSD_ACCOUNTS_FILE_NAME, NULL);
	if (!accounts_file_path)
		return;

	if (!accounts || !accounts->len) {
		unlink(accounts_file_path);
		goto out;
	}

	accounts_file_fd = g_open(accounts_file_path, O_CREAT | O_WRONLY | O_SYNC | O_TRUNC, S_IRUSR | S_IWUSR);
	if (-1 == accounts_file_fd)
		goto out;

	accounts_file = fdopen(accounts_file_fd, "w");
	if (!accounts_file)
		goto out;

	for (n = 0; n < accounts->len; n++) {
		account = (struct MAPIAccount*)g_ptr_array_index(accounts, n);
		if (!account)
			continue;

		// mailbox
		fputs(account->mailbox, accounts_file);
		fputs("\n", accounts_file);

		// exchange or domino server name
		fputs(account->server, accounts_file);
		fputs("\n", accounts_file);

		// profile name
		fputs(account->profile, accounts_file);
		fputs("\n", accounts_file);
	}

out:
	g_free(accounts_file_path);

	if (accounts_file) {
		fflush(accounts_file);
		fclose(accounts_file);
	} else if (-1 != accounts_file_fd)
		close(accounts_file_fd);
}

/* size of internal read buffer - maximum length of a mailbox or profile name */
#define ACCOUNT_ENTRY_SIZE 300

/* Will read all accounts from disk */
static void
brutus_read_accounts(GPtrArray **accounts)
{
	int c = 1; // anything but '\n'
	int count = 0;
	struct MAPIAccount *account = NULL;
	char buf[ACCOUNT_ENTRY_SIZE+1] = { '\0' };
	FILE *accounts_file = NULL;
	int accounts_file_fd = -1;
	gchar *accounts_file_path = g_strconcat(getenv("HOME"), "/", BRUTUSD_ACCOUNTS_FILE_NAME, NULL);
	if (!accounts_file_path)
		goto err;

	if (*accounts)
		brutus_free_accounts(*accounts);
	*accounts = NULL;

	accounts_file_fd = g_open(accounts_file_path, O_CREAT | O_RDONLY | O_SYNC , S_IRUSR | S_IWUSR);
	g_free(accounts_file_path);
	if (-1 == accounts_file_fd)
		goto err;

	accounts_file = fdopen(accounts_file_fd, "r");
	if (!accounts_file) {
		close(accounts_file_fd);
		goto err;
	}

	// count entries
	while (c != EOF) {
		if ('\n' == c)
			count++;
		c = getc(accounts_file);
	}
	if (3 > count) // return if too few
		goto err;
	{ // check for divisibility with 3
		div_t q = div(count, 3);
		if (q.rem)
			goto err;
		else
			count = q.quot;
	}
	rewind(accounts_file);

	*accounts = g_ptr_array_sized_new(count);
	if (!*accounts)
		goto out;

	/* 256 comes from 255+\n+\0 */
	/* 65 comes from 64+\n+\0 */
	while (count--) {
		account = g_new(struct MAPIAccount, 1);
		if (!account)
			goto err;

		// mailbox
		if (!fgets(buf, ACCOUNT_ENTRY_SIZE, accounts_file))
			goto err;
		if (256 <= strlen(buf)) // corrupt mailbox name.
			goto err;
		buf[strlen(buf)-1] = '\0'; // we know the last character is \n so we terminate the string at that
		account->mailbox = CORBA_string_dup(buf);
		memset((void*)buf, '\0', ACCOUNT_ENTRY_SIZE);

		// exchange server
		if (!fgets(buf, ACCOUNT_ENTRY_SIZE, accounts_file))
			goto err;
		if (256 <= strlen(buf)) // corrupt server name.
			goto err;
		buf[strlen(buf)-1] = '\0';
		account->server = CORBA_string_dup(buf);
		memset((void*)buf, '\0', ACCOUNT_ENTRY_SIZE);

		// profile name
		if (!fgets(buf, ACCOUNT_ENTRY_SIZE, accounts_file))
			goto err;
		if (65 <= strlen(buf)) // corrupt profile name
			goto err;
		buf[strlen(buf)-1] = '\0';
		account->profile = CORBA_string_dup(buf);
		memset((void*)buf, '\0', ACCOUNT_ENTRY_SIZE);

		account->session = CORBA_OBJECT_NIL;
		account->instance_id = 0;

		g_ptr_array_add(*accounts, (gpointer)account);
	}

out:
	fclose(accounts_file);
	return;

err:
	brutus_free_accounts(*accounts);
	*accounts = g_ptr_array_new();

	if (accounts_file)
		fclose(accounts_file);
	brutus_store_accounts(NULL);

	return;
}

/*** App-specific servant structures ***/

#if !defined(_typedef_impl_POA_BRUTUS_BrutusProxy_)
#define _typedef_impl_POA_BRUTUS_BrutusProxy_ 1
typedef struct {
	POA_BRUTUS_BrutusProxy servant;
	PortableServer_POA poa;

	/* ------ add private attributes here ------ */
	gboolean using_lorica;
	Lorica_ReferenceMapper mapper;
	GPtrArray *accounts;
	GArray *local_objects_map;
	CORBA_unsigned_long accumulated_wrap_count;
	BRUTUS_BrutusProxy_QueryState query_state;
	/* ------ ---------- end ------------ ------ */

} impl_POA_BRUTUS_BrutusProxy;
#endif


/*** Implementation stub prototypes ***/

#if !defined(_decl_impl_BRUTUS_BrutusProxy__destroy_)
#define _decl_impl_BRUTUS_BrutusProxy__destroy_ 1
static void
impl_BRUTUS_BrutusProxy__destroy(impl_POA_BRUTUS_BrutusProxy *servant,
				 CORBA_Environment *ev);
#endif

#if !defined(_decl_impl_BRUTUS_BrutusProxy_wrapObject_)
#define _decl_impl_BRUTUS_BrutusProxy_wrapObject_ 1
static CORBA_unsigned_long
impl_BRUTUS_BrutusProxy_wrapObject(impl_POA_BRUTUS_BrutusProxy *servant,
				   const BRUTUS_BrutusTC type_code,
				   const CORBA_Object filling,
				   CORBA_Object* wrapper,
				   CORBA_Environment *ev);
#endif

#if !defined(_decl_impl_BRUTUS_BrutusProxy_killWrapper_)
#define _decl_impl_BRUTUS_BrutusProxy_killWrapper_ 1
static void
impl_BRUTUS_BrutusProxy_killWrapper(impl_POA_BRUTUS_BrutusProxy *servant,
				    const CORBA_unsigned_long Key,
				    CORBA_Environment *ev);
#endif

#if !defined(_decl_impl_BRUTUS_BrutusProxy_getProfile_)
#define _decl_impl_BRUTUS_BrutusProxy_getProfile_ 1
static void
impl_BRUTUS_BrutusProxy_getProfile(impl_POA_BRUTUS_BrutusProxy *servant,
				   const CORBA_char * mailbox,
				   const CORBA_char * server,
				   CORBA_string* profile,
				   CORBA_Environment *ev);
#endif

#if !defined(_decl_impl_BRUTUS_BrutusProxy_querySession_)
#define _decl_impl_BRUTUS_BrutusProxy_querySession_ 1
static BRUTUS_BrutusProxy_QueryState
impl_BRUTUS_BrutusProxy_querySession(impl_POA_BRUTUS_BrutusProxy *servant,
				     const CORBA_char * mailbox,
				     const CORBA_char * server,
				     BRUTUS_IMAPISession* session,
				     CORBA_Environment *ev);
#endif

#if !defined(_decl_impl_BRUTUS_BrutusProxy_offerSession_)
#define _decl_impl_BRUTUS_BrutusProxy_offerSession_ 1
static void
impl_BRUTUS_BrutusProxy_offerSession(impl_POA_BRUTUS_BrutusProxy *servant,
				     const BRUTUS_IMAPISession session,
				     const CORBA_char * mailbox,
				     const CORBA_char * server,
				     const CORBA_char * profile,
				     CORBA_Environment *ev);
#endif


/*** epv structures ***/

#if !defined(_impl_BRUTUS_BrutusProxy_base_epv_)
#define _impl_BRUTUS_BrutusProxy_base_epv_ 1
static PortableServer_ServantBase__epv impl_BRUTUS_BrutusProxy_base_epv = {
	NULL,             /* _private data */
	(gpointer) & impl_BRUTUS_BrutusProxy__destroy, /* finalize routine */
	NULL,             /* default_POA routine */
};
#endif

#if !defined(_impl_BRUTUS_BrutusProxy_epv_)
#define _impl_BRUTUS_BrutusProxy_epv_ 1
static POA_BRUTUS_BrutusProxy__epv impl_BRUTUS_BrutusProxy_epv = {
	NULL, /* _private */
	(gpointer)&impl_BRUTUS_BrutusProxy_wrapObject,
	(gpointer)&impl_BRUTUS_BrutusProxy_killWrapper,
	(gpointer)&impl_BRUTUS_BrutusProxy_getProfile,
	(gpointer)&impl_BRUTUS_BrutusProxy_querySession,
	(gpointer)&impl_BRUTUS_BrutusProxy_offerSession,
};
#endif


/*** vepv structures ***/

#if !defined(_impl_BRUTUS_BrutusProxy_vepv_)
#define _impl_BRUTUS_BrutusProxy_vepv_ 1
static POA_BRUTUS_BrutusProxy__vepv impl_BRUTUS_BrutusProxy_vepv = {
	&impl_BRUTUS_BrutusProxy_base_epv,
	&impl_BRUTUS_BrutusProxy_epv,
};
#endif


/*** Stub implementations ***/

#if !defined(_impl_BRUTUS_BrutusProxy__create_)
#define _impl_BRUTUS_BrutusProxy__create_ 1
static BRUTUS_BrutusProxy impl_BRUTUS_BrutusProxy__create(PortableServer_POA poa,
							  Lorica_ReferenceMapper mapper,
							  CORBA_Environment *ev)
{
	CORBA_boolean cb;
	BRUTUS_BrutusProxy retval;
	impl_POA_BRUTUS_BrutusProxy *newservant;
	PortableServer_ObjectId *objid;

	newservant = g_new0(impl_POA_BRUTUS_BrutusProxy, 1);
	newservant->servant.vepv = &impl_BRUTUS_BrutusProxy_vepv;
	newservant->poa = (PortableServer_POA) CORBA_Object_duplicate((CORBA_Object)poa, ev);
	POA_BRUTUS_BrutusProxy__init((PortableServer_Servant)newservant, ev);
	/* Before servant is going to be activated all
	 * private attributes must be initialized.  */

	/* ------ init private attributes here ------ */

	/* lorica */
	cb = CORBA_Object_is_nil((CORBA_Object)mapper, ev);
	if (ORBIT2_EX(ev) || cb) {
	no_lorica:
		newservant->using_lorica = FALSE;
		newservant->mapper = CORBA_OBJECT_NIL;
	} else {
		newservant->using_lorica = TRUE;
		newservant->mapper = CORBA_Object_duplicate((CORBA_Object)mapper, ev);
		if (ORBIT2_EX(ev))
			goto no_lorica;
	}

	/* initialize state and locking */
	newservant->query_state = BRUTUS_BrutusProxy_idle;

	/* read all accounts */
	newservant->accounts = NULL;
	brutus_read_accounts(&newservant->accounts);

	/* store the local wrappers for later deallocation */
	newservant->local_objects_map = g_array_new(TRUE, FALSE, sizeof(gpointer));
	newservant->accumulated_wrap_count = 1;
	/* ------ ---------- end ------------- ------ */

	objid = PortableServer_POA_activate_object(poa, newservant, ev);
	CORBA_free(objid);
	retval = PortableServer_POA_servant_to_reference(poa, newservant, ev);

	return retval;
}
#endif

#if !defined(_impl_BRUTUS_BrutusProxy__destroy_)
#define _impl_BRUTUS_BrutusProxy__destroy_ 1
static void
impl_BRUTUS_BrutusProxy__destroy(impl_POA_BRUTUS_BrutusProxy *servant, CORBA_Environment *ev)
{
	guint idx = 0;
	struct WrapperItem *item = NULL;

	CORBA_Object_release((CORBA_Object)servant->poa, ev);

	/* No further remote method calls are delegated to
	 * servant and you may free your private attributes. */
	/* ------ free private attributes here ------ */

	CORBA_Object_release((CORBA_Object)servant->mapper, ev);

	brutus_free_accounts(servant->accounts);

	do {
		if (item)
			g_free(item);

		item = g_array_index(servant->local_objects_map, struct WrapperItem*, idx++);
		if (!item)
			break;

		CORBA_Object_release(item->local_object, ev);
		if (ORBIT2_EX(ev))
			continue;

	} while (item);
	g_array_free(servant->local_objects_map, TRUE);

	/* ------ ---------- end ------------- ------ */

	POA_BRUTUS_BrutusProxy__fini((PortableServer_Servant)servant, ev);

	g_free (servant);
}
#endif

#if !defined(_impl_BRUTUS_BrutusProxy_wrapObject_)
#define _impl_BRUTUS_BrutusProxy_wrapObject_ 1
static CORBA_unsigned_long
impl_BRUTUS_BrutusProxy_wrapObject(impl_POA_BRUTUS_BrutusProxy *servant,
				   const BRUTUS_BrutusTC type_code,
				   const CORBA_Object filling,
				   CORBA_Object *wrapper,
				   CORBA_Environment *ev)
{
	PortableServer_ObjectId *objid = NULL;
	struct WrapperItem *item = NULL;
	CORBA_unsigned_long retval = 0;

	*wrapper = CORBA_OBJECT_NIL;

	if (servant->using_lorica) {
		*wrapper = Lorica_ReferenceMapper_as_server(servant->mapper,
							    filling,
							    "",
							    CORBA_OBJECT_NIL,
							    ev);
		if (ORBIT2_EX(ev)) {
			kill(0, SIGUSR2); // restart
			return 0;
		}
	} else {
		switch (type_code) {
		case BRUTUS_BrutusCheck_tc :
			*wrapper = (CORBA_Object)impl_BRUTUS_BrutusCheck__create((BRUTUS_BrutusCheck)filling, servant->poa, ev);
			break;
		case BRUTUS_IMAPIAdviseSink_tc :
			*wrapper = (CORBA_Object)impl_BRUTUS_IMAPIAdviseSink__create((BRUTUS_IMAPIAdviseSink)filling, servant->poa, ev);
			break;
		case BRUTUS_SRestrictionContainer_tc :
			*wrapper = (CORBA_Object)impl_BRUTUS_SRestrictionContainer__create((BRUTUS_SRestrictionContainer)filling, servant->poa, ev);
			break;
		case BRUTUS_SAndRestriction_tc :
			*wrapper = (CORBA_Object)impl_BRUTUS_SAndRestriction__create((BRUTUS_SAndRestriction)filling, servant->poa, ev);
			break;
		case BRUTUS_SOrRestriction_tc :
			*wrapper = (CORBA_Object)impl_BRUTUS_SOrRestriction__create((BRUTUS_SOrRestriction)filling, servant->poa, ev);
			break;
		case BRUTUS_SNotRestriction_tc :
			*wrapper = (CORBA_Object)impl_BRUTUS_SNotRestriction__create((BRUTUS_SNotRestriction)filling, servant->poa, ev);
			break;
		case BRUTUS_SContentRestriction_tc :
			*wrapper = (CORBA_Object)impl_BRUTUS_SContentRestriction__create((BRUTUS_SContentRestriction)filling, servant->poa, ev);
			break;
		case BRUTUS_SPropertyRestriction_tc :
			*wrapper = (CORBA_Object)impl_BRUTUS_SPropertyRestriction__create((BRUTUS_SPropertyRestriction)filling, servant->poa, ev);
			break;
		case BRUTUS_SComparePropsRestriction_tc :
			*wrapper = (CORBA_Object)impl_BRUTUS_SComparePropsRestriction__create((BRUTUS_SComparePropsRestriction)filling, servant->poa, ev);
			break;
		case BRUTUS_SBitMaskRestriction_tc :
			*wrapper = (CORBA_Object)impl_BRUTUS_SBitMaskRestriction__create((BRUTUS_SBitMaskRestriction)filling, servant->poa, ev);
			break;
		case BRUTUS_SSizeRestriction_tc :
			*wrapper = (CORBA_Object)impl_BRUTUS_SSizeRestriction__create((BRUTUS_SSizeRestriction)filling, servant->poa, ev);
			break;
		case BRUTUS_SExistRestriction_tc :
			*wrapper = (CORBA_Object)impl_BRUTUS_SExistRestriction__create((BRUTUS_SExistRestriction)filling, servant->poa, ev);
			break;
		case BRUTUS_SSubRestriction_tc :
			*wrapper = (CORBA_Object)impl_BRUTUS_SSubRestriction__create((BRUTUS_SSubRestriction)filling, servant->poa, ev);
			break;
		case BRUTUS_SCommentRestriction_tc :
			*wrapper = (CORBA_Object)impl_BRUTUS_SCommentRestriction__create((BRUTUS_SCommentRestriction)filling, servant->poa, ev);
			break;
		case BRUTUS_IUnknown_tc :
			*wrapper = (CORBA_Object)impl_BRUTUS_IUnknown__create((BRUTUS_IUnknown)filling, servant->poa, ev);
			break;
		case BRUTUS_BrutusLogOn_tc :
		case BRUTUS_BrutusSession_tc :
		case BRUTUS_IABContainer_tc :
		case BRUTUS_IABLogon_tc :
		case BRUTUS_IABProvider_tc :
		case BRUTUS_IAddrBook_tc :
		case BRUTUS_IAttach_tc :
		case BRUTUS_IDistList_tc :
		case BRUTUS_IExchangeExportChanges_tc :
		case BRUTUS_IExchangeFavorites_tc :
		case BRUTUS_IExchangeImportContentsChanges_tc :
		case BRUTUS_IExchangeImportHierarchyChanges_tc :
		case BRUTUS_IExchangeManageStore_tc :
		case BRUTUS_IExchangeModifyTable_tc :
		case BRUTUS_IExchangeRuleAction_tc :
		case BRUTUS_IMAPIContainer_tc :
		case BRUTUS_IMAPIControl_tc :
		case BRUTUS_IMAPIFolder_tc :
		case BRUTUS_IMAPIForm_tc :
		case BRUTUS_IMAPIFormAdviseSink_tc :
		case BRUTUS_IMAPIFormContainer_tc :
		case BRUTUS_IMAPIFormFactory_tc :
		case BRUTUS_IMAPIFormInfo_tc :
		case BRUTUS_IMAPIFormMgr_tc :
		case BRUTUS_IMAPIMessageSite_tc :
		case BRUTUS_IMAPIProgress_tc :
		case BRUTUS_IMAPIProp_tc :
		case BRUTUS_IMAPISession_tc :
		case BRUTUS_IMAPIStatus_tc :
		case BRUTUS_IMAPISupport_tc :
		case BRUTUS_IMAPITable_tc :
		case BRUTUS_IMAPIViewAdviseSink_tc :
		case BRUTUS_IMAPIViewContext_tc :
		case BRUTUS_IMSLogon_tc :
		case BRUTUS_IMSProvider_tc :
		case BRUTUS_IMailUser_tc :
		case BRUTUS_IMessage_tc :
		case BRUTUS_IMsgServiceAdmin_tc :
		case BRUTUS_IMsgStore_tc :
		case BRUTUS_IPersistMessage_tc :
		case BRUTUS_IProfAdmin_tc :
		case BRUTUS_IProfSect_tc :
		case BRUTUS_IPropData_tc :
		case BRUTUS_IProviderAdmin_tc :
		case BRUTUS_ISequentialStream_tc :
		case BRUTUS_ISpoolerHook_tc :
		case BRUTUS_IStream_tc :
		case BRUTUS_ITableData_tc :
		case BRUTUS_ITnef_tc :
		case BRUTUS_IXPLogon_tc :
		case BRUTUS_IXPProvider_tc :
		case BRUTUS_log_server_tc :
			CORBA_exception_set_system(ev, ex_CORBA_NO_IMPLEMENT, CORBA_COMPLETED_NO);
			break;
		default :
			CORBA_exception_set_system(ev, ex_CORBA_BAD_PARAM, CORBA_COMPLETED_NO);
			break;
		};

		if (ORBIT2_EX(ev)) {
			*wrapper = CORBA_OBJECT_NIL;
			return 0;
		}
	}

	item = g_new(struct WrapperItem, 1);
	if (!item) {
		if (servant->using_lorica) {
			Lorica_ReferenceMapper_remove_server(servant->mapper,
							     *wrapper,
							     ev);
		}  else {
			objid = PortableServer_POA_reference_to_id(servant->poa, *wrapper, ev);
			if (ORBIT2_EX(ev)) {
				if (objid)
					CORBA_free(objid);
				goto out;
			}
			if (!objid)
				goto out;

			PortableServer_POA_deactivate_object(servant->poa, objid, ev);
			CORBA_free(objid);
		}
		*wrapper = CORBA_OBJECT_NIL;

	out:
		CORBA_exception_set_system(ev, ex_CORBA_NO_MEMORY, CORBA_COMPLETED_NO);
		return 0;
	}

	while (!g_static_mutex_trylock(&map_mutex))
		g_usleep(10);

	item->key = servant->accumulated_wrap_count++;

	item->local_object = CORBA_Object_duplicate(*wrapper, ev);
	g_array_append_val(servant->local_objects_map, item);
	g_static_mutex_unlock(&map_mutex);

	retval = item->key;

	return retval;
}
#endif

#if !defined(_impl_BRUTUS_BrutusProxy_killWrapper_)
#define _impl_BRUTUS_BrutusProxy_killWrapper_ 1
static void
impl_BRUTUS_BrutusProxy_killWrapper(impl_POA_BRUTUS_BrutusProxy *servant,
				    const CORBA_unsigned_long Key,
				    CORBA_Environment *ev)
{
	PortableServer_ObjectId *objid = NULL;
	struct WrapperItem *item = NULL;
	guint idx = 0;

	while (!g_static_mutex_trylock(&map_mutex))
		g_usleep(10);
	do {
		item = g_array_index(servant->local_objects_map, struct WrapperItem*, idx);
		if (!item)
			break;
		if (Key == item->key)
			break;
		idx++;
	} while (item);
	if (!item)
		goto out;

	if (servant->using_lorica) {
		Lorica_ReferenceMapper_remove_server(servant->mapper,
						     item->local_object,
						     ev);
		if (ORBIT2_EX(ev))
			goto out;
	} else {
		objid = PortableServer_POA_reference_to_id(servant->poa, item->local_object, ev);
		if (ORBIT2_EX(ev)) {
			if (objid)
				CORBA_free(objid);
			goto out;
		}
		if (!objid)
			goto out;

		PortableServer_POA_deactivate_object(servant->poa, objid, ev);
		CORBA_free(objid);
		if (ORBIT2_EX(ev))
			goto out;
	}

	CORBA_Object_release(item->local_object, ev);
	if (ORBIT2_EX(ev))
		goto out;

out:
	if (item) {
		g_free(item);
		g_array_remove_index(servant->local_objects_map, idx);
	}
	g_static_mutex_unlock(&map_mutex);

	/* ------ ---------- end ------------ ------ */
}
#endif

#if !defined(_impl_BRUTUS_BrutusProxy_getProfile_)
#define _impl_BRUTUS_BrutusProxy_getProfile_ 1
static void
impl_BRUTUS_BrutusProxy_getProfile(impl_POA_BRUTUS_BrutusProxy *servant,
				   const CORBA_char * mailbox,
				   const CORBA_char * server,
				   CORBA_string* profile,
				   CORBA_Environment *ev)
{
	guint n;

	if (!g_static_mutex_trylock(&st_mutex)) {
		CORBA_exception_set_system(ev, ex_CORBA_NO_PERMISSION, CORBA_COMPLETED_NO);
		return;
	}

	n = 0;
	if (brutus_account_find_match(servant->accounts, mailbox, server, &n)) {
		struct MAPIAccount *account = g_ptr_array_index(servant->accounts, n);
		if (!account)
			goto inv;
		*profile = CORBA_string_dup(account->profile);
	} else {
	inv:
		*profile = CORBA_string_dup("");
	}

	g_static_mutex_unlock(&st_mutex);
}
#endif

#if !defined(_impl_BRUTUS_BrutusProxy_querySession_)
#define _impl_BRUTUS_BrutusProxy_querySession_ 1
static BRUTUS_BrutusProxy_QueryState
impl_BRUTUS_BrutusProxy_querySession(impl_POA_BRUTUS_BrutusProxy *servant,
				     const CORBA_char *mailbox,
				     const CORBA_char *server,
				     BRUTUS_IMAPISession *session,
				     CORBA_Environment *ev)
{
	BRUTUS_BrutusProxy_QueryState retv;
	guint n = 0;

	*session = CORBA_OBJECT_NIL;

	if (!g_static_mutex_trylock(&st_mutex)) {
		CORBA_exception_set_system(ev, ex_CORBA_NO_PERMISSION, CORBA_COMPLETED_NO);
		return servant->query_state;
	}

	if (BRUTUS_BrutusProxy_offerInProgress == servant->query_state)
		retv = BRUTUS_BrutusProxy_offerInProgress;
	else
		retv = BRUTUS_BrutusProxy_mustOffer;

	if (BRUTUS_BrutusProxy_idle != servant->query_state)
		goto out;

	if (brutus_account_find_session(servant->accounts, mailbox, server, &n)) {
		struct MAPIAccount *account = (struct MAPIAccount*)g_ptr_array_index(servant->accounts, n);
		if (!account)
			goto inv;
		*session = (BRUTUS_IMAPISession)CORBA_Object_duplicate((CORBA_Object)account->session, ev);

		retv = BRUTUS_BrutusProxy_queryOK;
		servant->query_state = BRUTUS_BrutusProxy_idle;
	} else {
	inv:
		retv = BRUTUS_BrutusProxy_mustOffer;
		servant->query_state = BRUTUS_BrutusProxy_offerInProgress;
	}

out:
	g_static_mutex_unlock(&st_mutex);

	return retv;
}
#endif

#if !defined(_impl_BRUTUS_BrutusProxy_offerSession_)
#define _impl_BRUTUS_BrutusProxy_offerSession_ 1
static void
impl_BRUTUS_BrutusProxy_offerSession(impl_POA_BRUTUS_BrutusProxy *servant,
				     const BRUTUS_IMAPISession session,
				     const CORBA_char * mailbox,
				     const CORBA_char * server,
				     const CORBA_char * profile,
				     CORBA_Environment *ev)
{
	guint n;
	CORBA_boolean cb;
	struct MAPIAccount *account;

	if (!g_static_mutex_trylock(&st_mutex)) {
		CORBA_exception_set_system(ev, ex_CORBA_NO_PERMISSION, CORBA_COMPLETED_NO);
		return;
	}

	// will force all clients to query before offering
	if (BRUTUS_BrutusProxy_offerInProgress != servant->query_state)
		goto out;
	servant->query_state = BRUTUS_BrutusProxy_idle;

	/* check for NIL offer and go to idle if so */
	cb = CORBA_Object_is_nil((CORBA_Object)session, ev);
	if (ORBIT2_EX(ev) || cb)
		goto out;

	/* check for valid offer - go to idle if not */
	BRUTUS_IMAPISession_ping(session, ev);
	if (ORBIT2_EX(ev)) // offer is invalid
		goto out;

	n = 0;
	account = NULL;
	if (brutus_account_find_match(servant->accounts, mailbox, server, &n)) {
		account = g_ptr_array_index(servant->accounts, n);
		if (!account)
			goto inv;

		// do not overwrite an existing and valid entry
		cb = CORBA_Object_is_nil((CORBA_Object)account->session, ev);
		if (!ORBIT2_EX(ev) && !cb) {
			BRUTUS_IMAPISession_ping(account->session, ev);
			if (!ORBIT2_EX(ev))
				goto out;
			else {
				CORBA_Object_release(account->session, ev);
				account->session = CORBA_OBJECT_NIL;
			}
		}

		// force a profile update
		if (account->profile)
			CORBA_free(account->profile);
		account->profile = CORBA_string_dup(profile);
		account->session = (BRUTUS_IMAPISession)CORBA_Object_duplicate(session, ev);
		if (ORBIT2_EX(ev)) {
			account->session = CORBA_OBJECT_NIL;
			brutus_store_accounts(servant->accounts);
			goto out;
		}
	} else { // new account
	inv:
		account = g_new(struct MAPIAccount, 1);
		if (!account)
			goto out;

		account->session = (BRUTUS_IMAPISession)CORBA_Object_duplicate(session, ev);
		if (ORBIT2_EX(ev))
			account->session = CORBA_OBJECT_NIL;

		account->mailbox = CORBA_string_dup(mailbox);
		account->server = CORBA_string_dup(server);
		account->profile = CORBA_string_dup(profile);

		g_ptr_array_add(servant->accounts, account);
	}
	brutus_store_accounts(servant->accounts);

out:
	g_static_mutex_unlock(&st_mutex);
}
#endif
