/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */

/*
 *    Brutus proxy test program.
 *    Copyright (C) 2007 OMC Denmark ApS.
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 *    MA 02111-1307 USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include <glib/gstdio.h>

#include "IMAPISession.h"

#include "brutusd.h"
#include "brutus-proxy.h"


#ifndef ORBIT2_EX
#define ORBIT2_EX(_ev_) ((NULL != _ev_) && ((_ev_)->_major != CORBA_NO_EXCEPTION))
#endif

#ifndef PRINT_ORBIT2_EX
#define PRINT_ORBIT2_EX(_ev_)                                                     \
        G_STMT_START {                                                            \
		switch ((_ev_)->_major) {				          \
		case CORBA_NO_EXCEPTION :				          \
			g_print("No exception.\n");                               \
			break;						          \
		case CORBA_USER_EXCEPTION :				          \
			g_print("%s@%d: User Exception : %s\n",                   \
				__FILE__, __LINE__, CORBA_exception_id ((_ev_))); \
			break;						          \
		case CORBA_SYSTEM_EXCEPTION :				          \
			g_print("%s@%d: System Exception : %s\n",                 \
				__FILE__, __LINE__, CORBA_exception_id ((_ev_))); \
			break;						          \
		default :						          \
			break;						          \
		}							          \
	} G_STMT_END
#endif

static CORBA_ORB
create_orb(void)
{
	CORBA_ORB orb = CORBA_OBJECT_NIL;
	CORBA_Environment ev[1];
	int init_argc = 1;
	char **init_argv = (char**)malloc(sizeof(char*) * (init_argc));
	if (!init_argv)
		return CORBA_OBJECT_NIL;

	CORBA_exception_init(ev);

	init_argv[0] = strdup("DUMMY_ARGV0");

	// initialize the ORB
	orb = CORBA_ORB_init(&init_argc, init_argv, "brutus-query_orb_orbit-io-thread", ev);
	if (ORBIT2_EX(ev))
		orb = CORBA_OBJECT_NIL;

	free(init_argv[0]);
	free(init_argv);

	CORBA_exception_free(ev);

	return orb;
}

static gboolean
wait_for_daemon(const int fd,
		unsigned int attempts)
{
	struct flock lock = {
		.l_type = F_WRLCK,
		.l_start = 0,
		.l_whence = SEEK_SET,
		.l_len = 0,
	};

	if (-1 == fd)
		return FALSE;

	while (attempts) {
		if (fcntl(fd, F_GETLK, &lock))
			return FALSE;

		if (F_UNLCK == lock.l_type)
			return TRUE;

		/* sleep for a while (0.2 seconds) and try once more */
		g_usleep(200000);

		attempts--;
	}

	return FALSE;
}

static CORBA_Object
brutus_object_ref_from_file(CORBA_ORB orb,
			    const gchar *filename,
			    CORBA_Environment *ev)
{
	CORBA_Object obj = CORBA_OBJECT_NIL;
	CORBA_char *objref = NULL;
	FILE *file = NULL;
	int fd = -1;
	int attempts = 10;
	struct stat st;
	size_t c = 0;

	file = g_fopen(filename, "r");
	if (!file)
		return CORBA_OBJECT_NIL;
	fd = fileno(file);
	if (-1 == fd)
		goto out;

	/* 
	 * We assume that the daemon holds an exclusive write lock 
	 * on the ior file to indicate that it is ready to be used 
	 */
	if (!wait_for_daemon(fd, attempts))
		goto out;

	if (g_stat(filename, &st))
		goto out;

	objref = g_malloc0(st.st_size);
	if (!objref)
		goto out;

	// must work even if sizeof(char) != sizeof(CORBA_char)
	c = fread((void*)objref, sizeof(CORBA_char), (size_t)(st.st_size/sizeof(CORBA_char)), file);
	if (c != (size_t)(st.st_size/sizeof(CORBA_char)))
		goto out;

	obj = (CORBA_Object)CORBA_ORB_string_to_object(orb, objref, ev);

out:
	if (objref)
		g_free(objref);
	fclose(file);

	return obj;
}

static BRUTUS_BrutusProxy
get_proxy_ref(CORBA_ORB orb)
{
	char *home_dir = getenv("HOME");
	gchar *ref_path = NULL;
	BRUTUS_BrutusProxy proxy = CORBA_OBJECT_NIL;
	CORBA_boolean cb = CORBA_FALSE;
	CORBA_Environment ev[1];

	CORBA_exception_init(ev);

	ref_path = g_strconcat(home_dir, "/", BRUTUSD_REF_FILE_NAME, NULL);
	if (!ref_path)
		goto out;
	proxy = (BRUTUS_BrutusProxy)brutus_object_ref_from_file(orb, ref_path, ev);
	g_free(ref_path);
	if (ORBIT2_EX(ev))
		goto out;

	cb = CORBA_Object_is_nil((CORBA_Object)proxy, ev);
	if (ORBIT2_EX(ev)) 
		goto out;

	if (cb) {
		g_printf("Could not get the Brutus proxy reference\n");
		goto out;
	}

out:
	CORBA_exception_free(ev);

	return proxy;

}

static void print_legalese(const char *prog_name)
{
	g_print("\n");
	g_print("                      -*- DISCLAIMER START -*-\n");
	g_print("\n");
	g_print("  * %s - A sample program to test the Brutus proxy daemon.\n", prog_name);
	g_print("  * Copyright (C) 2007 OMC Denmark ApS\n");
	g_print("  * \n");
	g_print("  * This program is free software; you can redistribute it and/or modify\n");
	g_print("  * it under the terms of the GNU General Public License as published by\n");
	g_print("  * the Free Software Foundation; either version 2 of the License, or\n");
	g_print("  * (at your option) any later version.\n");
	g_print("  * \n");
	g_print("  * This program is distributed in the hope that it will be useful,\n");
	g_print("  * but WITHOUT ANY WARRANTY; without even the implied warranty of\n");
	g_print("  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the\n");
	g_print("  * GNU General Public License for more details.\n");
	g_print("  * \n");
	g_print("  * You should have received a copy of the GNU General Public License\n");
	g_print("  * along with this program; if not, write to the Free Software\n");
	g_print("  * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.\n");
	g_print("\n");
	g_print("                      -*- DISCLAIMER END -*-\n");
	g_print("\n");
}
int
main(int argc, char *argv[])
{
	int retv = EXIT_FAILURE;
	BRUTUS_IMAPISession mapi_session = CORBA_OBJECT_NIL;
	BRUTUS_BrutusProxy_QueryState query_state = BRUTUS_BrutusProxy_idle;
	BRUTUS_BrutusProxy proxy = CORBA_OBJECT_NIL;
	CORBA_boolean cb = CORBA_TRUE;
	CORBA_ORB orb = CORBA_OBJECT_NIL;
	gboolean release = FALSE;
	CORBA_Environment ev[1];

	g_type_init();
	g_thread_init(NULL);

	print_legalese(argv[0]);

	CORBA_exception_init(ev);

	orb = create_orb();
	proxy = get_proxy_ref(orb);
	cb = CORBA_Object_is_nil((CORBA_Object)proxy, ev);
	if (ORBIT2_EX(ev)) 
		goto out;
	else if (cb) {
		g_print("The Brutus Proxy daemon is not running\n");
		goto out;
	}

	do {
		query_state = BRUTUS_BrutusProxy_querySession(proxy,
							      "Bogus_Non_Existing_Mailbox",
							      "bogus.test",
							      &mapi_session,
							      ev);
		if (ORBIT2_EX(ev)) {
			if (!strcmp(CORBA_exception_id(ev), ex_CORBA_NO_PERMISSION)) {
				g_usleep(100000); // sleep for 0.1 second
				continue;
			}
			PRINT_ORBIT2_EX(ev);
			goto out;
		}
	} while (FALSE);

	switch (query_state) {
	case BRUTUS_BrutusProxy_mustOffer :
		g_print("BRUTUS_BrutusProxy_mustOffer\n");
		do {
			BRUTUS_BrutusProxy_offerSession(proxy,
							CORBA_OBJECT_NIL,
							"",
							"",
							"",
							ev);
			if (ORBIT2_EX(ev)) {
				if (!strcmp(CORBA_exception_id(ev), ex_CORBA_NO_PERMISSION)) {
					g_usleep(100000); // sleep for 0.1 second
					continue;
				}
				PRINT_ORBIT2_EX(ev);
			}
		} while (FALSE);
		break;
	case BRUTUS_BrutusProxy_offerInProgress :
		g_print("BRUTUS_BrutusProxy_offerInProgress");
		break;
	case BRUTUS_BrutusProxy_queryOK : // got session
		g_print("??? BRUTUS_BrutusProxy_queryOK ???");
		release = TRUE;
		break;
	default:
		g_print("Unknown query response");
	}


	goto out;

out:
	CORBA_Object_release((CORBA_Object)proxy, ev);
	if (release)
		CORBA_Object_release((CORBA_Object)mapi_session, ev);

	CORBA_ORB_shutdown(orb, FALSE, ev);
	CORBA_ORB_destroy(orb, ev);
	CORBA_Object_release((CORBA_Object)orb, ev);

	CORBA_exception_free(ev);

	return retv;
}
